# Copyright 2024 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import io
import os
from collections import OrderedDict
from collections import defaultdict
import datetime
import numpy as np
import pandas as pd
import math
import re
import logging
from jinja2 import Template
from statistics import median
from cami_amber import plots
from cami_amber import binning_classes
from cami_amber.utils import labels as utils_labels
from cami_amber.utils import load_ncbi_taxinfo

import matplotlib
matplotlib.use('Agg')
import seaborn as sns
import matplotlib.pyplot as pltx
from matplotlib.colors import rgb2hex
from matplotlib.colors import Normalize

from version import __version__

from bokeh.plotting import figure
from bokeh.layouts import column, row
from bokeh.models.widgets import TableColumn, Div, Select
from bokeh.models import TabPanel, Tabs
from bokeh.models import (DataTable,
                          CustomJS,
                          Legend, LegendItem,
                          Band,
                          CustomJSTickFormatter,
                          HoverTool)
from bokeh.models.tickers import FixedTicker
from bokeh.resources import INLINE
from bokeh.embed import file_html
from bokeh.plotting import ColumnDataSource


AVG_OVER_SAMPLES = '[average over samples]'
CLICK_ON_LEGENDS_DIV = '<div>' + AVG_OVER_SAMPLES + '</div><div style="padding-top: 10px;">Click on the legends to enable/disable the results for a tool.</div>'

TEMPLATE = Template('''
{% from macros import embed %}
<!DOCTYPE html>
<html lang="en">
  {% block head %}
  <head>
  {% block inner_head %}
    <meta charset="utf-8">
    <title>{% block title %}{{ title | e if title else "Bokeh Plot" }}{% endblock %}</title>
  {%  block preamble -%}{%- endblock %}
  {%  block resources %}
    <style>
      html, body {
        box-sizing: border-box;
        height: 100%;
        margin: 8px;
        padding: 0;
      }
      .bk-fit-content {width: fit-content; width: -moz-fit-content;}
      .bk-width-auto {width: auto !important; height: auto !important;}
      .bk-display-block {display: block !important;}
      .bk-float-left {float: left;}
      bk-width-auto-main>div {width: -webkit-fill-available !important;}
      div.bk-width-auto-main {width: -webkit-fill-available !important;}
      .bk-tabs-margin{margin-top: 20px !important;}
      .bk-tabs-margin-lr{margin-left: 10px; margin-right: 10px}
      .bk-root {display: flex; justify-content: center;}
      .bk-padding-top {padding-top: 10px;}
      .bk-padding-top2 {padding-top: 20px;}
      .bk-combo-box > div:first-child {
        width: auto !important;
        padding-right: 40px;
      }
    </style>
  {%   block css_resources -%}
    {{- bokeh_css if bokeh_css }}
  {%-  endblock css_resources %}
  {%   block js_resources -%}
    {{  bokeh_js if bokeh_js }}
  {%-  endblock js_resources %}
  {%  endblock resources %}
  {%  block postamble %}{% endblock %}
  {% endblock inner_head %}
  </head>
  {% endblock head%}
  {% block body %}
  <body>
  {%  block inner_body %}
  {%    block contents %}
  {%      for doc in docs %}
  {{        embed(doc) if doc.elementid }}
  {%-       for root in doc.roots %}
  {%          block root scoped %}
  {{            embed(root) }}
  {%          endblock %}
  {%        endfor %}
  {%      endfor %}
  {%    endblock contents %}
  {{ plot_script | indent(4) }}
  {%  endblock inner_body %}
  </body>
  {% endblock body%}
</html>
''')


def create_heatmap_bar(output_dir):
    fig = pltx.figure(figsize=(2, 2))
    x = np.arange(25).reshape(5, 5)
    cm = sns.diverging_palette(12, 250, sep=30, l=50, n=15, s=90, as_cmap=True)
    ax = sns.heatmap(x, cmap=cm, cbar=False)
    cbar = pltx.colorbar(ax.get_children()[0], orientation='horizontal')

    # hide border
    cbar.outline.set_visible(False)

    # hide ticks
    cbar.set_ticks([])

    # hide heatmap
    pltx.gca().set_visible(False)

    fig.savefig(os.path.join(output_dir, 'heatmap_bar.png'), dpi=100, format='png', bbox_inches='tight', pad_inches=-.001, transparent=True)
    pltx.close(fig)


class MidpointNormalize(Normalize):
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))


def get_colors_and_ranges(name, all_values):
    color1 = 'dodgerblue'
    color2 = 'red'
    hue1 = 12
    hue2 = 240

    metrics = [utils_labels.MISCLASSIFICATION_PER_BP, utils_labels.MISCLASSIFICATION_PER_SEQ]
    if name in metrics + [x + utils_labels.UNFILTERED for x in metrics]:
        return color2, color1, hue2, hue1, 0, 1
    metrics = [utils_labels.UNIFRAC_BP, utils_labels.UNIFRAC_SEQ]
    if name in metrics + [x + utils_labels.UNFILTERED for x in metrics]:
        return color2, color1, hue2, hue1, 0, max(all_values)
    return color1, color2, hue1, hue2, 0, 1


def get_heatmap_colors(pd_series, **args):
    values = pd_series.tolist()

    metrics = [utils_labels.AVG_PRECISION_BP_SEM, utils_labels.AVG_RECALL_BP_SEM, utils_labels.AVG_PRECISION_SEQ_SEM,
               utils_labels.AVG_RECALL_SEQ_SEM, utils_labels.AVG_RECALL_BP_SEM_CAMI1, utils_labels.AVG_RECALL_SEQ_SEM_CAMI1]
    if pd_series.name in metrics + [x + utils_labels.UNFILTERED for x in metrics]:
        return ['background-color: white' for x in values]

    dropped_gs = False
    if pd_series.index[0] == utils_labels.GS:
        values = values[1:]
        dropped_gs = True
    if len(values) == 0:
        return ['']

    notnan_values = [x for x in values if isinstance(x, (float, int)) and not np.isnan(x)]
    if not notnan_values:
        red = 'background-color: red'
        return [red for x in values] if not dropped_gs else [''] + [red for x in values]

    color1, color2, hue1, hue2, min_value, max_value = get_colors_and_ranges(pd_series.name, values)

    cm = sns.diverging_palette(hue1, hue2, sep=50, l=80, n=15, s=90, as_cmap=True)
    norm = MidpointNormalize(vmin=min_value, vmax=max_value, midpoint=median(notnan_values))

    normed = norm([round(x, 3) if not math.isnan(x) else max_value for x in values])
    heatmap_colors = [rgb2hex(x) for x in pltx.cm.get_cmap(cm)(normed)]
    return_colors = []
    for val, x in zip(values, heatmap_colors):
        if val <= min_value:
            return_colors.append('background-color: {}'. format(color2))
        elif val >= max_value:
            return_colors.append('background-color: {}'. format(color1))
        elif math.isnan(val):
            return_colors.append('background-color: red')
        else:
            return_colors.append('background-color: {}'. format(x))

    if dropped_gs:
        return [''] + return_colors
    else:
        return return_colors


def create_title_div(id, name, info):
    div = Div(text="""<div style="text-align:left;font-size: 20pt;font-weight: bold;">{1}<span style="float: right;font-size: 10pt;font-weight:normal;">{2}</span>""".format(id, name, info), styles={'width': 'auto'})
    div.stylesheets = ["""
    .bk-clearfix {
      width: -webkit-fill-available;
      width: -moz-available;
    }
    """]
    return div


def create_metrics_per_bin_panel(pd_bins, bins_columns, sample_ids_list, output_dir, binning_type):
    styles = [{'selector': 'td', 'props': [('width', '92pt')]},
              {'selector': 'th', 'props': [('width', '92pt'), ('text-align', 'left')]},
              {'selector': 'expand-toggle:checked ~ * .data', 'props': [('background-color', 'white !important')]}]

    tools = pd_bins[utils_labels.TOOL].unique().tolist()
    tool_to_sample_to_html = defaultdict(list)

    for tool, pd_tool_groupby in pd_bins.groupby(utils_labels.TOOL):
        pd_tool_sample_groupby = pd_tool_groupby.groupby('sample_id')
        available_samples = pd_tool_sample_groupby.groups.keys()
        for sample_id in sample_ids_list:
            if sample_id not in available_samples:
                tool_to_sample_to_html[tool].append('')
                continue
            pd_tool_sample = pd_tool_sample_groupby.get_group(sample_id)
            pd_tool_sample = pd_tool_sample[list(bins_columns.keys())].rename(columns=dict(bins_columns))
            if binning_type == 'taxonomic':
                pd_tool_sample['Taxon ID'] = pd_tool_sample['Taxon ID'].astype('Int64')
            tool_sample_html = pd_tool_sample.head(500).style.set_table_styles(styles).format(precision=3).hide(axis='index').to_html()
            tool_sample_html += '<div style="padding-top: 20px; padding-bottom: 20px;">{}</div>'.format('Complete table available in: ' + os.path.join(output_dir, binning_type, tool, 'metrics_per_bin.tsv'))
            tool_to_sample_to_html[tool].append(tool_sample_html)
    tool_to_sample_to_html['all_samples'] = sample_ids_list

    table_div = Div(text="""<div>{}</div>""".format(tool_to_sample_to_html[tools[0]][0]))

    source = ColumnDataSource(data=tool_to_sample_to_html)

    select_tool = Select(title="Binner:", value=tools[0], options=tools)
    select_sample = Select(title="Sample:", value='0', options=list(zip(map(str, range(len(sample_ids_list))), sample_ids_list)))
    select_tool_sample_callback = CustomJS(args=dict(source=source), code="""
        select_sample.options = [];
        const options_array = [];
        for (let index in source.data[select_tool.value]) {
            if (source.data[select_tool.value][index] != "") {
                options_array.push([index, source.data["all_samples"][index]])
            }
        }
        select_sample.options = options_array;
        if (source.data[select_tool.value][select_sample.value] != "") {
            mytable.text = source.data[select_tool.value][select_sample.value];
        } else {
            mytable.text = source.data[select_tool.value][options_array[0][0]];
        }
    """)
    select_tool.js_on_change('value', select_tool_sample_callback)
    select_sample.js_on_change('value', select_tool_sample_callback)
    select_tool_sample_callback.args["mytable"] = table_div
    select_tool_sample_callback.args["select_tool"] = select_tool
    select_tool_sample_callback.args["select_sample"] = select_sample

    table_column = row(column(row(select_tool, select_sample), table_div, sizing_mode='scale_width'), sizing_mode='scale_width')
    metrics_bins_panel = TabPanel(child=table_column, title="Metrics per bin")
    return metrics_bins_panel


def create_contamination_completeness_table(pd_bins, min_completeness, max_contamination):
    df = binning_classes.GenomeQuery.calc_num_recovered_genomes(pd_bins, min_completeness, max_contamination)

    def create_table_column(field):
        return TableColumn(title=field, field=field, width=600)

    dt = DataTable(source=ColumnDataSource(df),
                   columns=list(map(lambda x: create_table_column(x), df.columns.values)),
                   width=800,
                   height=1000,
                   reorderable=True,
                   selectable=True)
    return [column(dt)]


def create_heatmap_div():
    heatmap_legend = '<img src="heatmap_bar.png" /><div style="text-align:left;font-size: 11px;">Worst<span style="float:right;">Best</span><span style="margin-right: 36px;float:right;">Median</span></div>'
    heatmap_legend_div = Div(text=heatmap_legend, styles={"width": "155px", "margin-bottom": "-10px"})
    return heatmap_legend_div


def get_labels_genome():
    return ([(utils_labels.ACCURACY_PER_BP, utils_labels.TOOLTIP_ACCURACY_PER_BP),
             (utils_labels.ACCURACY_PER_SEQ, utils_labels.TOOLTIP_ACCURACY_PER_SEQ),
             (utils_labels.MISCLASSIFICATION_PER_BP, utils_labels.TOOLTIP_MISCLASSIFICATION_PER_BP),
             (utils_labels.MISCLASSIFICATION_PER_SEQ, utils_labels.TOOLTIP_MISCLASSIFICATION_PER_SEQ),
             (utils_labels.PRECISION_PER_BP, utils_labels.TOOLTIP_PRECISION_PER_BP),
             (utils_labels.PRECISION_PER_SEQ, utils_labels.TOOLTIP_PRECISION_PER_SEQ),
             (utils_labels.RECALL_PER_BP, utils_labels.TOOLTIP_RECALL_PER_BP),
             (utils_labels.RECALL_PER_SEQ, utils_labels.TOOLTIP_RECALL_PER_SEQ),
             (utils_labels.F1_SCORE_PER_BP, utils_labels.TOOLTIP_F1_SCORE_PER_BP),
             (utils_labels.F1_SCORE_PER_SEQ, utils_labels.TOOLTIP_F1_SCORE_PER_SEQ),
             (utils_labels.RI_BY_BP, utils_labels.TOOLTIP_RI_BY_BP),
             (utils_labels.RI_BY_SEQ, utils_labels.TOOLTIP_RI_BY_SEQ),
             (utils_labels.ARI_BY_BP, utils_labels.TOOLTIP_ARI_BY_BP),
             (utils_labels.ARI_BY_SEQ, utils_labels.TOOLTIP_ARI_BY_SEQ),
             (utils_labels.AVG_PRECISION_BP, utils_labels.TOOLTIP_AVG_PRECISION_BP),
             (utils_labels.AVG_PRECISION_BP_SEM, utils_labels.TOOLTIP_AVG_PRECISION_BP_SEM),
             (utils_labels.AVG_PRECISION_SEQ, utils_labels.TOOLTIP_AVG_PRECISION_SEQ),
             (utils_labels.AVG_PRECISION_SEQ_SEM, utils_labels.TOOLTIP_AVG_PRECISION_SEQ_SEM),
             (utils_labels.AVG_RECALL_BP, utils_labels.TOOLTIP_AVG_RECALL_BP),
             (utils_labels.AVG_RECALL_BP_CAMI1, utils_labels.TOOLTIP_AVG_RECALL_BP_CAMI1),
             (utils_labels.AVG_RECALL_BP_SEM, utils_labels.TOOLTIP_AVG_RECALL_BP_SEM),
             (utils_labels.AVG_RECALL_BP_SEM_CAMI1, utils_labels.TOOLTIP_AVG_RECALL_BP_SEM),
             (utils_labels.AVG_RECALL_SEQ, utils_labels.TOOLTIP_AVG_RECALL_SEQ),
             (utils_labels.AVG_RECALL_SEQ_CAMI1, utils_labels.TOOLTIP_AVG_RECALL_SEQ_CAMI1),
             (utils_labels.F1_SCORE_BP, utils_labels.TOOLTIP_F1_SCORE_BP),
             (utils_labels.F1_SCORE_SEQ, utils_labels.TOOLTIP_F1_SCORE_SEQ),
             (utils_labels.F1_SCORE_BP_CAMI1, utils_labels.TOOLTIP_F1_SCORE_BP),
             (utils_labels.F1_SCORE_SEQ_CAMI1, utils_labels.TOOLTIP_F1_SCORE_SEQ),
             (utils_labels.AVG_RECALL_SEQ_SEM, utils_labels.TOOLTIP_AVG_RECALL_SEQ_SEM),
             (utils_labels.AVG_RECALL_SEQ_SEM_CAMI1, utils_labels.TOOLTIP_AVG_RECALL_SEQ_SEM),
             (utils_labels.PERCENTAGE_ASSIGNED_BPS, utils_labels.TOOLTIP_PERCENTAGE_ASSIGNED_BPS),
             (utils_labels.PERCENTAGE_ASSIGNED_SEQS, utils_labels.TOOLTIP_PERCENTAGE_ASSIGNED_SEQS),
             (utils_labels.UNIFRAC_BP, utils_labels.TOOLTIP_UNIFRAC_BP),
             (utils_labels.UNIFRAC_SEQ, utils_labels.TOOLTIP_UNIFRAC_SEQ)])


def get_labels_taxonomic():
    return ([(utils_labels.ACCURACY_PER_BP, utils_labels.TOOLTIP_ACCURACY_PER_BP_TAX),
             (utils_labels.ACCURACY_PER_SEQ, utils_labels.TOOLTIP_ACCURACY_PER_SEQ_TAX),
             (utils_labels.MISCLASSIFICATION_PER_BP, utils_labels.TOOLTIP_MISCLASSIFICATION_PER_BP),
             (utils_labels.MISCLASSIFICATION_PER_SEQ, utils_labels.TOOLTIP_MISCLASSIFICATION_PER_SEQ),
             (utils_labels.PRECISION_PER_BP, utils_labels.TOOLTIP_PRECISION_PER_BP_TAX),
             (utils_labels.PRECISION_PER_SEQ, utils_labels.TOOLTIP_PRECISION_PER_SEQ_TAX),
             (utils_labels.RECALL_PER_BP, utils_labels.TOOLTIP_RECALL_PER_BP_TAX),
             (utils_labels.RECALL_PER_SEQ, utils_labels.TOOLTIP_RECALL_PER_SEQ_TAX),
             (utils_labels.F1_SCORE_PER_BP, utils_labels.TOOLTIP_F1_SCORE_PER_BP),
             (utils_labels.F1_SCORE_PER_SEQ, utils_labels.TOOLTIP_F1_SCORE_PER_SEQ),
             (utils_labels.RI_BY_BP, utils_labels.TOOLTIP_RI_BY_BP_TAX),
             (utils_labels.RI_BY_SEQ, utils_labels.TOOLTIP_RI_BY_SEQ_TAX),
             (utils_labels.ARI_BY_BP, utils_labels.TOOLTIP_ARI_BY_BP),
             (utils_labels.ARI_BY_SEQ, utils_labels.TOOLTIP_ARI_BY_SEQ),
             (utils_labels.AVG_PRECISION_BP, utils_labels.TOOLTIP_AVG_PRECISION_BP_TAX),
             (utils_labels.AVG_PRECISION_BP_SEM, utils_labels.TOOLTIP_AVG_PRECISION_BP_SEM),
             (utils_labels.AVG_PRECISION_SEQ, utils_labels.TOOLTIP_AVG_PRECISION_SEQ_TAX),
             (utils_labels.AVG_PRECISION_SEQ_SEM, utils_labels.TOOLTIP_AVG_PRECISION_SEQ_SEM),
             (utils_labels.AVG_RECALL_BP, utils_labels.TOOLTIP_AVG_RECALL_BP_TAX),
             (utils_labels.AVG_RECALL_BP_CAMI1, utils_labels.TOOLTIP_AVG_RECALL_BP_CAMI1),
             (utils_labels.AVG_RECALL_BP_SEM, utils_labels.TOOLTIP_AVG_RECALL_BP_SEM),
             (utils_labels.AVG_RECALL_SEQ, utils_labels.TOOLTIP_AVG_RECALL_SEQ_TAX),
             (utils_labels.AVG_RECALL_SEQ_CAMI1, utils_labels.TOOLTIP_AVG_RECALL_SEQ_CAMI1),
             (utils_labels.F1_SCORE_BP, utils_labels.TOOLTIP_F1_SCORE_BP),
             (utils_labels.F1_SCORE_SEQ, utils_labels.TOOLTIP_F1_SCORE_SEQ),
             (utils_labels.AVG_RECALL_SEQ_SEM, utils_labels.TOOLTIP_AVG_RECALL_SEQ_SEM),
             (utils_labels.PERCENTAGE_ASSIGNED_BPS, utils_labels.TOOLTIP_PERCENTAGE_ASSIGNED_BPS),
             (utils_labels.PERCENTAGE_ASSIGNED_SEQS, utils_labels.TOOLTIP_PERCENTAGE_ASSIGNED_SEQS),
             (utils_labels.UNIFRAC_BP, utils_labels.TOOLTIP_UNIFRAC_BP),
             (utils_labels.UNIFRAC_SEQ, utils_labels.TOOLTIP_UNIFRAC_SEQ)])


def create_table_html(df_summary, is_taxonomic=False, include_cami1=False):
    metrics1 = [utils_labels.AVG_PRECISION_BP,
                utils_labels.AVG_PRECISION_SEQ,
                utils_labels.AVG_RECALL_BP,
                utils_labels.AVG_RECALL_SEQ,
                utils_labels.F1_SCORE_BP,
                utils_labels.F1_SCORE_SEQ,
                utils_labels.AVG_PRECISION_BP_SEM,
                utils_labels.AVG_PRECISION_SEQ_SEM,
                utils_labels.AVG_RECALL_BP_SEM,
                utils_labels.AVG_RECALL_SEQ_SEM]

    if is_taxonomic and utils_labels.AVG_PRECISION_BP + utils_labels.UNFILTERED in df_summary.index:
        metrics1 += [x + utils_labels.UNFILTERED for x in metrics1]

    if include_cami1:
        metrics1 += [utils_labels.AVG_RECALL_BP_CAMI1,
                     utils_labels.AVG_RECALL_SEQ_CAMI1,
                     utils_labels.F1_SCORE_BP_CAMI1,
                     utils_labels.F1_SCORE_SEQ_CAMI1,
                     utils_labels.AVG_RECALL_BP_SEM_CAMI1,
                     utils_labels.AVG_RECALL_SEQ_SEM_CAMI1]
    metrics2 = [utils_labels.ACCURACY_PER_BP,
                utils_labels.ACCURACY_PER_SEQ,
                utils_labels.MISCLASSIFICATION_PER_BP,
                utils_labels.MISCLASSIFICATION_PER_SEQ,
                utils_labels.PRECISION_PER_BP,
                utils_labels.PRECISION_PER_SEQ,
                utils_labels.RECALL_PER_BP,
                utils_labels.RECALL_PER_SEQ,
                utils_labels.F1_SCORE_PER_BP,
                utils_labels.F1_SCORE_PER_SEQ,
                utils_labels.RI_BY_BP,
                utils_labels.RI_BY_SEQ,
                utils_labels.ARI_BY_BP,
                utils_labels.ARI_BY_SEQ,
                utils_labels.PERCENTAGE_ASSIGNED_BPS,
                utils_labels.PERCENTAGE_ASSIGNED_SEQS]
    if is_taxonomic:
        metrics2.append(utils_labels.UNIFRAC_BP)
        metrics2.append(utils_labels.UNIFRAC_SEQ)

    if is_taxonomic and utils_labels.AVG_PRECISION_BP + utils_labels.UNFILTERED in df_summary.index:
        metrics2 += [x + utils_labels.UNFILTERED for x in metrics2]

    all_metrics = [metrics1, metrics2]
    metrics1_label = utils_labels.QUALITY_OF_BINS
    metrics2_label = utils_labels.QUALITY_OF_SAMPLE
    all_metrics_labels = [metrics1_label, metrics2_label]

    styles = [{'selector': 'td', 'props': [('width', '115pt')]},
              {'selector': 'th', 'props': [('width', '115pt'), ('text-align', 'left')]},
              {'selector': 'th:nth-child(1)', 'props': [('width', '210pt'), ('font-weight', 'normal')]},
              {'selector': '', 'props': [('width', 'max-content'), ('width', '-moz-max-content'), ('border-top', '1px solid lightgray'), ('border-spacing', '0px')]},
              {'selector': 'expand-toggle:checked ~ * .data', 'props': [('background-color', 'white !important')]}]
    styles_hidden_thead = styles + [{'selector': 'thead', 'props': [('display', 'none')]}]

    def get_html_dict(metrics):
        d_dict = {}
        for tuple in metrics:
            d_dict[tuple[0] + '<'] = '<div class="tooltip">{}<span class="tooltiptext">{}: {}</span></div><'.format(utils_labels.LABELS[tuple[0]], utils_labels.LABELS[tuple[0]], tuple[1])
            d_dict[tuple[0] + utils_labels.UNFILTERED + '<'] = '<div class="tooltip">{}<span class="tooltiptext">{}: {}</span></div><'.format(utils_labels.LABELS[tuple[0]] + utils_labels.UNFILTERED, utils_labels.LABELS[tuple[0]], tuple[1])
        return d_dict

    d = get_html_dict(get_labels_taxonomic() if is_taxonomic else get_labels_genome())

    pattern = re.compile('|'.join(map(re.escape, d)))

    def translate(match):
        return d[match.group(0)]

    df_summary.index.name = None

    html = io.StringIO()
    first_metrics = True
    for metrics, metrics_label in zip(all_metrics, all_metrics_labels):
        html.write('<p style="margin-bottom: auto"><b>{}</b></p>'.format(metrics_label))
        df_metrics = df_summary.loc[metrics]
        sorted_columns = df_metrics.columns.tolist()
        df_metrics = df_metrics.loc[:, sorted_columns]

        if first_metrics:
            this_style = styles
            first_metrics = False
        else:
            this_style = styles_hidden_thead
        html.write(df_metrics.style.apply(get_heatmap_colors, df_metrics=df_metrics, axis=1).format(
            precision=3).set_table_styles(this_style).to_html())
    html = html.getvalue()
    html = pattern.sub(translate, html)

    tooltips = """
    <style>
        .tooltip {
            position: relative;
            display: inline-block;
            border-bottom: 1px dashed lightgray;
            cursor: help;
        }
        div.tooltip span.tooltiptext {
            visibility: hidden;
            width: 280px;
            background-color: #555;
            color: #fff;
            text-align: center;
            border-radius: 6px;
            padding: 5px 0;
            position: absolute;
            z-index: 1;
            bottom: 125%;
            left: 50%;
            margin-left: -60px;
            opacity: 0;
        }
        div.tooltip span.tooltiptext::after {
            content: "";
            position: absolute;
            top: 100%;
            left: 50%;
            margin-left: -5px;
            border-width: 5px;
            border-style: solid;
            border-color: #555 transparent transparent transparent;
        }
        div.tooltip:hover span.tooltiptext {
            visibility: visible;
            opacity: 1;
        }
        tr:hover {
            outline: 1px solid black;
        }
    </style>"""

    return '{}<div style="margin-bottom:10pt;">{}</div>'.format(tooltips, html)


def create_precision_recall_figure(df_summary, xname1, yname1, xname2, yname2, title):
    colors_list = plots.create_colors_list()
    bokeh_colors = [matplotlib.colors.to_hex(c) for c in colors_list]

    legend_it = []
    tooltips1 = [(utils_labels.TOOL, '@index'),
                 (xname1, '@{' + xname1 + '}'),
                 (yname1, '@{' + yname1 + '}')]
    tooltips2 = [(utils_labels.TOOL, '@index'),
                 (xname2, '@{' + xname2 + '}'),
                 (yname2, '@{' + yname2 + '}')]
    p = figure(title=title, width=580, height=400, x_range=(0, 1), y_range=(0, 1), toolbar_location="below")
    p.xaxis.axis_label = utils_labels.LABELS[xname1]
    p.yaxis.axis_label = utils_labels.LABELS[yname1]
    p.xaxis.axis_label_text_font_style = 'normal'
    p.yaxis.axis_label_text_font_style = 'normal'
    for color, (index, row) in zip(bokeh_colors, df_summary.iterrows()):
        tool = index
        source = ColumnDataSource(data=row.to_frame().T)
        pcircle = p.scatter(xname1, yname1, source=source, color=color, fill_alpha=0.2, size=10)
        px = p.scatter(xname2, yname2, marker='x', source=source, color=color, size=10)
        p.add_tools(HoverTool(tooltips=tooltips1, renderers=[pcircle], visible=False))
        p.add_tools(HoverTool(tooltips=tooltips2, renderers=[px], visible=False))
        legend_it.append(LegendItem(label=tool, renderers=[pcircle, px]))
    pcircle = p.scatter([0], [0], color='black', fill_alpha=0, size=0)
    legend_it.append(LegendItem(label='by bp', renderers=[pcircle]))
    px = p.scatter([0], [0], marker='x', color='black', fill_alpha=0, size=0)
    legend_it.append(LegendItem(label='by seq', renderers=[px]))

    p.add_layout(Legend(items=legend_it), 'right')
    p.legend.click_policy = 'hide'
    return p


def create_precision_recall_all_genomes_scatter(pd_genome_bins, tools):
    colors_list = plots.create_colors_list()
    bokeh_colors = [matplotlib.colors.to_hex(c) for c in colors_list]

    p = figure(title='Quality per bin', width=580, height=400, x_range=(0, 1), y_range=(0, 1), toolbar_location="below")
    p.add_tools(HoverTool(tooltips=[('Sample', '@sample_id'),
                                    ('Genome', '@mapping_id'),
                                    ('Purity of bin (bp)', '@precision_bp'),
                                    ('Completeness of bin (bp)', '@recall_bp')], visible=False))

    legend_it = []
    for color, tool in zip(bokeh_colors, tools):
        source = ColumnDataSource(data=pd_genome_bins[pd_genome_bins[utils_labels.TOOL] == tool])
        pcircle = p.scatter('precision_bp', 'recall_bp', color=color, alpha=0.8, source=source)
        legend_it.append((tool, [pcircle]))
    p.add_layout(Legend(items=legend_it), 'right')
    p.xaxis.axis_label = 'Purity per bin (bp)'
    p.yaxis.axis_label = 'Completeness per bin (bp)'
    p.xaxis.axis_label_text_font_style = 'normal'
    p.yaxis.axis_label_text_font_style = 'normal'
    p.legend.click_policy = 'hide'
    return p


def create_contamination_plot(pd_bins, tools, title, xlabel, ylabel, create_column_function):
    pd_bins_copy = pd_bins[[utils_labels.TOOL, 'precision_bp', 'recall_bp']].copy().dropna(subset=['precision_bp'])
    create_column_function(pd_bins_copy)

    colors_list = plots.create_colors_list()
    bokeh_colors = [matplotlib.colors.to_hex(c) for c in colors_list]

    p = figure(title=title, width=580, height=400, toolbar_location="below")
    p.x_range.start = 0
    # p.y_range.start = pd_bins_copy['newcolumn'].min()
    p.y_range.end = 1
    legend_it = []
    for color, tool in zip(bokeh_colors, tools):
        pd_tool_bins = pd_bins_copy[pd_bins_copy[utils_labels.TOOL] == tool]
        pd_tool_bins = pd_tool_bins.sort_values(by='newcolumn', ascending=False).reset_index()
        pd_tool_bins = pd_tool_bins.drop(['index'], axis=1)

        source = ColumnDataSource(data=pd_tool_bins)

        pline = p.line('index', 'newcolumn', color=color, line_width=2, source=source)
        legend_it.append((tool, [pline]))
    p.add_layout(Legend(items=legend_it), 'right')
    p.xaxis.axis_label = xlabel
    p.yaxis.axis_label = ylabel
    p.xaxis.axis_label_text_font_style = 'normal'
    p.yaxis.axis_label_text_font_style = 'normal'
    p.legend.click_policy = 'hide'
    return p


def create_tax_figure(tool, df_summary, metrics_list, errors_list):
    rank_indices = list(range(len(load_ncbi_taxinfo.RANKS)))

    df = df_summary.set_index("rank").reindex(load_ncbi_taxinfo.RANKS)
    df["x"] = rank_indices
    for metric, error in zip(metrics_list, errors_list):
        if error:
            df[metric + "upper"] = df[metric] + df[error]
            df[metric + "lower"] = df[metric] - df[error]
    source = ColumnDataSource(df.reset_index())

    p = figure(width=500, height=550, x_range=(0, 7), y_range=(0, 1))
    line_colors = ["#006cba", "#008000", "#ba9e00", "red"]
    legend_it = []
    for i, (metric, error, color) in enumerate(zip(metrics_list, errors_list, line_colors)):
        pline = p.line(x='x', y=metric, line_color=color, source=source, line_width=2)
        legend_it.append(LegendItem(label=utils_labels.LABELS[metric], renderers=[pline]))
        if error:
            band = Band(base='x', lower=metric + "lower", upper=metric + "upper", source=source, level='underlay',
                              fill_alpha=.3, line_width=1, line_color='black', fill_color=color)
            p.add_layout(band)
            callback = CustomJS(args = dict(band=band), code="""
                if (band.visible == false)
                    band.visible = true;
                else
                    band.visible = false; """)
            pline.js_on_change('visible', callback)
            if i > 1:
                pline.visible = False
                band.visible = False

    p.xaxis.ticker = FixedTicker(ticks=rank_indices)
    p.xaxis.formatter = CustomJSTickFormatter(code="""
        var mapping = {0: "superkingdom", 1: "phylum", 2: "class", 3: "order", 4: "family", 5: "genus", 6: "species", 7: "strain"};
        return mapping[tick];
    """)
    p.xaxis.major_label_orientation = 1.2
    p.add_layout(Legend(items=legend_it, location=(10, 10)), 'above')
    p.xgrid[0].grid_line_color=None
    p.ygrid[0].grid_line_alpha=0.5
    p.title.text = tool
    p.legend.click_policy = 'hide'
    return p


def create_rankings_table(df_summary, show_rank=False):
    columns = [utils_labels.AVG_PRECISION_BP,
               utils_labels.AVG_RECALL_BP,
               utils_labels.PRECISION_PER_BP,
               utils_labels.RECALL_PER_BP,
               utils_labels.ARI_BY_SEQ,
               utils_labels.ARI_BY_BP,
               utils_labels.PERCENTAGE_ASSIGNED_BPS,
               utils_labels.ACCURACY_PER_BP]
    if show_rank:
        columns.insert(0, utils_labels.RANK)
    labels_dict = utils_labels.LABELS.copy()
    labels_dict[utils_labels.RANK] = 'Taxonomic rank'
    pd_rankings = df_summary[columns].rename(columns=labels_dict).round(decimals=5).reset_index()

    def create_table_column(field):
        return TableColumn(title=field, field=field, width=100)

    dt = DataTable(source=ColumnDataSource(pd_rankings),
                   columns=list(map(lambda x: create_table_column(x), pd_rankings.columns.values)),
                   width=1500,
                   height=1000,
                   reorderable=True,
                   selectable=True)
    return [column(dt)]


def create_genome_binning_plots_panel(pd_bins, pd_mean):
    click_div = Div(text=CLICK_ON_LEGENDS_DIV, styles={"width": "500px", "margin-top": "15px; margin-bottom:5px;"})
    purity_completeness_plot = column(create_precision_recall_figure(pd_mean, utils_labels.AVG_PRECISION_BP, utils_labels.AVG_RECALL_BP, utils_labels.AVG_PRECISION_SEQ, utils_labels.AVG_RECALL_SEQ, utils_labels.QUALITY_OF_BINS))
    purity_recall_bp_plot = column(create_precision_recall_figure(pd_mean, utils_labels.PRECISION_PER_BP, utils_labels.RECALL_PER_BP, utils_labels.PRECISION_PER_SEQ, utils_labels.RECALL_PER_SEQ, utils_labels.QUALITY_OF_SAMPLE))

    all_samples_div = Div(text='<div style="padding-top: 20px;">All samples</div>', styles={"width": "500px", "margin-top": "15px; margin-bottom:5px;"})
    all_bins_plot = column(create_precision_recall_all_genomes_scatter(pd_bins, pd_mean.index.tolist()))
    completeness_contamination_plot = column(create_contamination_plot(pd_bins, pd_mean.index.tolist(), 'Completeness - contamination', 'Index of bin (sorted by completeness - contamination (bp))', 'Completeness - contamination (bp)', plots.create_completeness_minus_contamination_column))
    contamination_plot = column(create_contamination_plot(pd_bins, pd_mean.index.tolist(), 'Contamination', 'Index of bin (sorted by contamination (bp))', 'Contamination (bp)', plots.create_contamination_column))

    return TabPanel(child=column([click_div, purity_completeness_plot, purity_recall_bp_plot, all_samples_div, all_bins_plot, completeness_contamination_plot, contamination_plot], sizing_mode='scale_width'), title='Plots')


def create_genome_binning_html(df_summary, pd_bins, labels, sample_ids_list, options):
    if pd_bins.empty:
        return None
    df_summary_g = df_summary[df_summary[utils_labels.BINNING_TYPE] == 'genome']
    if df_summary_g.empty:
        return None

    sample_to_html = {}
    available_samples = list(df_summary_g[utils_labels.SAMPLE].unique())
    available_samples = [sample_id for sample_id in sample_ids_list if sample_id in available_samples]
    available_tools = list(df_summary_g[utils_labels.TOOL].unique())
    available_tools = [tool for tool in labels if tool in available_tools]

    pd_mean = df_summary_g.groupby(utils_labels.TOOL).mean(numeric_only=True).reindex(available_tools)
    pd_mean[utils_labels.SAMPLE] = AVG_OVER_SAMPLES
    sample_to_html[AVG_OVER_SAMPLES] = [create_table_html(pd_mean.T, include_cami1=True)]

    for sample_id, pd_sample in df_summary_g.groupby(utils_labels.SAMPLE):
        sample_to_html[sample_id] = [create_table_html(pd_sample.set_index(utils_labels.TOOL).T, include_cami1=True)]

    sample_ids_list_combo = [AVG_OVER_SAMPLES] + available_samples

    genome_div = Div(text="""<div style="margin-bottom:10pt;">{}</div>""".format(sample_to_html[sample_ids_list_combo[0]][0]))
    source = ColumnDataSource(data=sample_to_html)

    select_sample = Select(title="Sample:", value=sample_ids_list_combo[0], options=sample_ids_list_combo)
    select_sample_callback = CustomJS(args=dict(source=source), code="""
        mytable.text = source.data[select_sample.value][0];
    """)
    select_sample.js_on_change('value', select_sample_callback)
    select_sample_callback.args["mytable"] = genome_div
    select_sample_callback.args["select_sample"] = select_sample

    metrics_column = column(column(select_sample, create_heatmap_div(), genome_div, sizing_mode='scale_width'), sizing_mode='scale_width')
    metrics_panel = TabPanel(child=metrics_column, title="Metrics")

    plots_panel = create_genome_binning_plots_panel(pd_bins, pd_mean)

    bins_columns = utils_labels.get_genome_bins_columns()
    metrics_bins_panel = create_metrics_per_bin_panel(pd_bins, bins_columns, sample_ids_list, options.output_dir, 'genome')

    cc_table = create_contamination_completeness_table(pd_bins, options.min_completeness, options.max_contamination)
    cc_panel = TabPanel(child=row(cc_table), title="#Recovered genomes")

    rankings_panel = TabPanel(child=column([Div(text="Click on the columns header for sorting.", styles={"width": "500px", "margin-top": "20px"}),
                                        row(create_rankings_table(pd_mean.reset_index().set_index([utils_labels.SAMPLE, utils_labels.TOOL])))]), title="Rankings")

    tabs = Tabs(tabs=[metrics_panel, plots_panel, metrics_bins_panel, rankings_panel, cc_panel])

    return tabs


def create_plots_per_binner(df_summary_t):
    tools = df_summary_t[utils_labels.TOOL].unique().tolist()
    sample_div = Div(text=AVG_OVER_SAMPLES, styles={"margin-top": "15px; margin-bottom:5px;"})
    # using the same div breaks the layout
    sample_div2 = Div(text=AVG_OVER_SAMPLES, styles={"margin-top": "15px; margin-bottom:5px;"})

    metrics_list = [utils_labels.AVG_PRECISION_BP, utils_labels.AVG_RECALL_BP, utils_labels.AVG_PRECISION_SEQ, utils_labels.AVG_RECALL_SEQ]
    errors_list = [utils_labels.AVG_PRECISION_BP_SEM, utils_labels.AVG_RECALL_BP_SEM, utils_labels.AVG_PRECISION_SEQ_SEM, utils_labels.AVG_RECALL_SEQ_SEM]
    tools_figures = [create_tax_figure(tool, df_summary_t[df_summary_t[utils_labels.TOOL] == tool], metrics_list, errors_list) for tool in tools]
    tools_figures_columns = [column(x) for x in tools_figures]
    tools_column_unweighted = column([sample_div] + tools_figures_columns, sizing_mode='scale_width')

    metrics_list = [utils_labels.PRECISION_PER_BP, utils_labels.RECALL_PER_BP, utils_labels.PRECISION_PER_SEQ, utils_labels.RECALL_PER_SEQ]
    errors_list = ['', '', '', '']
    tools_figures_weighted = [create_tax_figure(tool, df_summary_t[df_summary_t[utils_labels.TOOL] == tool], metrics_list, errors_list) for tool in tools]
    tools_figures_columns = [column(x) for x in tools_figures_weighted]
    tools_column_weighted = column([sample_div2] + tools_figures_columns, sizing_mode='scale_width')

    tools_unweighted_panel = TabPanel(child=tools_column_unweighted, title=utils_labels.QUALITY_OF_BINS)
    tools_weighted_panel = TabPanel(child=tools_column_weighted, title=utils_labels.QUALITY_OF_SAMPLE)
    tools_tabs = Tabs(tabs=[tools_unweighted_panel, tools_weighted_panel])
    return tools_tabs


def create_tax_ranks_panel(qbins_plots_list, qsamples_plots_list, cc_plots_dict_list, contamination_plots_list):
    sample_div = Div(text=CLICK_ON_LEGENDS_DIV, styles={"width": "500px", "margin-top": "20px"})
    all_samples_div = Div(text='<div style="padding-top: 20px;">All samples</div>', styles={"width": "500px", "margin-top": "15px; margin-bottom:5px;"})
    all_samples_div2 = Div(text='<div style="padding-top: 20px;">All samples</div>', styles={"width": "500px", "margin-top": "15px; margin-bottom:5px;"})
    unweighted_panel = TabPanel(child=column([sample_div] + qbins_plots_list + [all_samples_div] + cc_plots_dict_list + [all_samples_div2] + contamination_plots_list, sizing_mode='scale_width'), title=utils_labels.QUALITY_OF_BINS)

    sample_div2 = Div(text=CLICK_ON_LEGENDS_DIV, styles={"width": "500px", "margin-top": "20px"})
    weighted_panel = TabPanel(child=column([sample_div2] + qsamples_plots_list, sizing_mode='scale_width'), title=utils_labels.QUALITY_OF_SAMPLE)
    tax_ranks_tabs = Tabs(tabs=[unweighted_panel, weighted_panel])
    return TabPanel(child=tax_ranks_tabs, title="Plots per taxonomic rank")


def create_taxonomic_binning_html(df_summary, pd_bins, labels, sample_ids_list, options):
    if pd_bins.empty:
        return None
    df_summary_t = df_summary[df_summary[utils_labels.BINNING_TYPE] == 'taxonomic'].copy()
    df_summary_t[[utils_labels.UNIFRAC_BP, utils_labels.UNIFRAC_SEQ]] = df_summary_t[[utils_labels.UNIFRAC_BP, utils_labels.UNIFRAC_SEQ]].apply(pd.to_numeric)

    rank_to_sample_to_html = defaultdict(list)
    qbins_plots_dict = OrderedDict([(rank, '') for rank in load_ncbi_taxinfo.RANKS])
    qsamples_plots_dict = OrderedDict([(rank, '') for rank in load_ncbi_taxinfo.RANKS])
    cc_plots_dict = OrderedDict([(rank, '') for rank in load_ncbi_taxinfo.RANKS])
    contamination_plots_dict = OrderedDict([(rank, '') for rank in load_ncbi_taxinfo.RANKS])

    pd_mean = df_summary_t.groupby([utils_labels.RANK, utils_labels.TOOL]).mean(numeric_only=True).reset_index()
    pd_mean[utils_labels.SAMPLE] = AVG_OVER_SAMPLES
    for rank, pd_mean_rank in pd_mean.groupby(utils_labels.RANK):
        available_tools = list(pd_mean_rank[utils_labels.TOOL].unique())
        available_tools = [tool for tool in labels if tool in available_tools]
        pd_mean_rank = pd_mean_rank.set_index(utils_labels.TOOL).reindex(available_tools)

        purity_completeness_plot = create_precision_recall_figure(pd_mean_rank, utils_labels.AVG_PRECISION_BP, utils_labels.AVG_RECALL_BP, utils_labels.AVG_PRECISION_SEQ, utils_labels.AVG_RECALL_SEQ, rank)
        purity_recall_bp_plot = create_precision_recall_figure(pd_mean_rank, utils_labels.PRECISION_PER_BP, utils_labels.RECALL_PER_BP, utils_labels.PRECISION_PER_SEQ, utils_labels.RECALL_PER_SEQ, rank)
        qbins_plots_dict[rank] = column(purity_completeness_plot)
        qsamples_plots_dict[rank] = column(purity_recall_bp_plot)

        pd_bins_rank = pd_bins[pd_bins['rank'] == rank]
        completeness_minus_contamination_plot = create_contamination_plot(pd_bins_rank, available_tools, rank + ' | Completeness - contamination', 'Index of bin (sorted by completeness - contamination (bp))', 'Completeness - contamination (bp)', plots.create_completeness_minus_contamination_column)
        contamination_plot = create_contamination_plot(pd_bins_rank, available_tools, rank + ' | Contamination', 'Index of bin (sorted by contamination (bp))', 'Contamination (bp)', plots.create_contamination_column)
        cc_plots_dict[rank] = column(completeness_minus_contamination_plot)
        contamination_plots_dict[rank] = column(contamination_plot)

        rank_to_sample_to_html[rank].append(create_table_html(pd_mean_rank.T, is_taxonomic=True))

    qbins_plots_list = [v for k, v in qbins_plots_dict.items() if v]
    qsamples_plots_list = [v for k, v in qsamples_plots_dict.items() if v]
    cc_plots_dict_list = [v for k, v in cc_plots_dict.items() if v]
    contamination_plots_list = [v for k, v in contamination_plots_dict.items() if v]

    pd_groupby_rank = df_summary_t.groupby(utils_labels.RANK)
    for rank, pd_rank in pd_groupby_rank:
        pd_rank_groupby_sample = pd_rank.groupby(utils_labels.SAMPLE)
        available_samples = pd_rank_groupby_sample.groups.keys()
        for sample_id in sample_ids_list:
            if sample_id not in available_samples:
                rank_to_sample_to_html[rank].append('')
                continue
            pd_rank_sample = pd_rank_groupby_sample.get_group(sample_id)
            pd_rank_sample = pd_rank_sample.set_index(utils_labels.TOOL).T
            rank_to_sample_to_html[rank].append(create_table_html(pd_rank_sample, is_taxonomic=True))

    if not rank_to_sample_to_html:
        return None

    sample_ids_list_combo = [AVG_OVER_SAMPLES] + sample_ids_list

    taxonomic_div = Div(text="""<div style="margin-bottom:10pt;">{}</div>""".format(rank_to_sample_to_html[load_ncbi_taxinfo.RANKS[0]][0]))
    source = ColumnDataSource(data=rank_to_sample_to_html)

    available_ranks = list(pd_groupby_rank.groups.keys())
    available_ranks_sorted = [rank for rank in load_ncbi_taxinfo.RANKS if rank in available_ranks]

    select_rank = Select(title="Taxonomic rank:", value=load_ncbi_taxinfo.RANKS[0], options=available_ranks_sorted)
    select_sample = Select(title="Sample:", value='0', options=list(zip(map(str, range(len(sample_ids_list_combo))), sample_ids_list_combo)))
    select_rank_sample_callback = CustomJS(args=dict(source=source), code="""
        mytable.text = source.data[select_rank.value][select_sample.value];
    """)
    select_rank.js_on_change('value', select_rank_sample_callback)
    select_sample.js_on_change('value', select_rank_sample_callback)
    select_rank_sample_callback.args["mytable"] = taxonomic_div
    select_rank_sample_callback.args["select_rank"] = select_rank
    select_rank_sample_callback.args["select_sample"] = select_sample

    metrics_column = column(column(row(select_sample, select_rank), create_heatmap_div(), taxonomic_div, sizing_mode='scale_width'), sizing_mode='scale_width')
    metrics_panel = TabPanel(child=metrics_column, title="Metrics")

    bins_columns = utils_labels.get_tax_bins_columns()
    if 'name' not in pd_bins.columns or pd_bins['name'].isnull().any():
        del bins_columns['name']

    metrics_bins_panel = create_metrics_per_bin_panel(pd_bins, bins_columns, sample_ids_list, options.output_dir, 'taxonomic')

    rankings_panel = TabPanel(child=column([Div(text="Click on the columns header for sorting.", styles={"width": "500px", "margin-top": "20px"}),
                                        row(create_rankings_table(pd_mean.reset_index().set_index([utils_labels.SAMPLE, utils_labels.TOOL]), True))]), title="Rankings")

    tax_ranks_panel = create_tax_ranks_panel(qbins_plots_list, qsamples_plots_list, cc_plots_dict_list, contamination_plots_list)

    tools_panel = TabPanel(child=create_plots_per_binner(pd_mean), title="Plots per binner")

    tabs = Tabs(tabs=[metrics_panel, tax_ranks_panel, tools_panel, metrics_bins_panel, rankings_panel])

    return tabs


def create_html(df_summary, pd_bins, labels, sample_ids_list, options, desc_text):
    logging.getLogger('amber').info('Creating HTML page')
    create_heatmap_bar(options.output_dir)
    tabs_list = []

    metrics_row_g = create_genome_binning_html(df_summary, pd_bins[pd_bins['rank'] == 'NA'], labels, sample_ids_list, options)
    if metrics_row_g:
        tabs_list.append(TabPanel(child=metrics_row_g, title="Genome binning"))

    metrics_row_t = create_taxonomic_binning_html(df_summary, pd_bins[pd_bins['rank'] != 'NA'], labels, sample_ids_list, options)
    if metrics_row_t:
        tabs_list.append(TabPanel(child=metrics_row_t, title="Taxonomic binning"))

    tabs = Tabs(tabs=tabs_list)

    title = create_title_div("main", "AMBER: Assessment of Metagenome BinnERs", " produced on {0} with AMBER version {1} ".format(
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M"), __version__))

    if desc_text:
        data_desc_div = Div(text="""<div style="text-align:left;font-size: 11pt;font-weight: bold;">{}""".format(desc_text))
        html_columns = column(title, data_desc_div, tabs, sizing_mode='scale_width')
    else:
        html_columns = column(title, tabs, sizing_mode='scale_width')

    html = file_html(models=html_columns, resources=INLINE, title='AMBER: Assessment of Metagenome BinnERs', template=TEMPLATE)

    with open(os.path.join(options.output_dir, "index.html"), 'w') as f:
        f.write(html)



from setuptools import setup, find_packages
from version import __version__
import glob


def dependencies():
    with open('requirements.txt', 'r') as f:
        return f.read().splitlines()


setup(
    name             = 'cami-amber',
    version          = __version__,
    description      = 'AMBER: Assessment of Metagenome BinnERs',
    author           = 'CAMI',
    author_email     = 'support@cami-challenge.org',
    url              = 'http://cami-challenge.org',
    license          = 'GNU General Public License v3 or later (GPLv3+)',
    scripts          = glob.glob('*.py'),
    install_requires = dependencies(),
    packages         = find_packages(),
    classifiers = [
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX'
    ]
)



# Copyright 2024 Department of Computational Biology for Infection Research - Helmholtz Centre for Infection Research
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import pandas as pd
import logging
import itertools
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('Agg')
import os
import math
from abc import ABC, abstractmethod
from collections import defaultdict
from collections import OrderedDict
from cami_amber.utils import labels as utils_labels
from cami_amber.utils import load_data
from cami_amber.utils import load_ncbi_taxinfo
from cami_amber.utils import ProfilingTools as pf
from cami_amber import unifrac_distance as uf
from cami_amber import precision_recall_per_bin
from cami_amber import plots


class Metrics:
    def __init__(self):
        self.__percentage_of_assigned_bps = .0
        self.__percentage_of_assigned_seqs = .0
        self.__accuracy_bp = .0
        self.__accuracy_seq = .0
        self.__rand_index_bp = .0
        self.__adjusted_rand_index_bp = .0
        self.__rand_index_seq = .0
        self.__adjusted_rand_index_seq = .0
        self.__precision_avg_bp = .0
        self.__precision_avg_bp_var = .0
        self.__precision_avg_bp_sem = .0
        self.__precision_avg_seq = .0
        self.__precision_avg_seq_sem = .0
        self.__precision_weighted_bp = .0
        self.__precision_weighted_seq = .0
        self.__recall_avg_bp = .0
        self.__recall_avg_bp_cami1 = .0
        self.__recall_avg_bp_var = .0
        self.__recall_avg_bp_var_cami1 = .0
        self.__recall_avg_bp_sem = .0
        self.__recall_avg_bp_sem_cami1 = .0
        self.__recall_avg_seq = .0
        self.__recall_avg_seq_cami1 = .0
        self.__recall_avg_seq_sem = .0
        self.__recall_avg_seq_sem_cami1 = .0
        self.__recall_weighted_bp = .0
        self.__recall_weighted_seq = .0
        self.__f1_score_bp = .0
        self.__f1_score_seq = .0
        self.__f1_score_bp_cami1 = .0
        self.__f1_score_seq_cami1 = .0
        self.__unifrac_bp = None
        self.__unifrac_seq = None

    @property
    def percentage_of_assigned_bps(self):
        return self.__percentage_of_assigned_bps

    @property
    def percentage_of_assigned_seqs(self):
        return self.__percentage_of_assigned_seqs

    @property
    def accuracy_bp(self):
        return self.__accuracy_bp

    @property
    def accuracy_seq(self):
        return self.__accuracy_seq

    @property
    def rand_index_bp(self):
        return self.__rand_index_bp

    @property
    def adjusted_rand_index_bp(self):
        return self.__adjusted_rand_index_bp

    @property
    def rand_index_seq(self):
        return self.__rand_index_seq

    @property
    def adjusted_rand_index_seq(self):
        return self.__adjusted_rand_index_seq

    @property
    def precision_avg_bp(self):
        return self.__precision_avg_bp

    @property
    def precision_avg_bp_var(self):
        return self.__precision_avg_bp_var

    @property
    def precision_avg_bp_sem(self):
        return self.__precision_avg_bp_sem

    @property
    def precision_avg_seq(self):
        return self.__precision_avg_seq

    @property
    def precision_avg_seq_sem(self):
        return self.__precision_avg_seq_sem

    @property
    def precision_weighted_bp(self):
        return self.__precision_weighted_bp

    @property
    def precision_weighted_seq(self):
        return self.__precision_weighted_seq

    @property
    def recall_avg_bp(self):
        return self.__recall_avg_bp

    @property
    def recall_avg_bp_cami1(self):
        return self.__recall_avg_bp_cami1

    @property
    def recall_avg_bp_var(self):
        return self.__recall_avg_bp_var

    @property
    def recall_avg_bp_var_cami1(self):
        return self.__recall_avg_bp_var_cami1

    @property
    def recall_avg_bp_sem(self):
        return self.__recall_avg_bp_sem

    @property
    def recall_avg_bp_sem_cami1(self):
        return self.__recall_avg_bp_sem_cami1

    @property
    def recall_avg_seq(self):
        return self.__recall_avg_seq

    @property
    def recall_avg_seq_cami1(self):
        return self.__recall_avg_seq_cami1

    @property
    def recall_avg_seq_sem(self):
        return self.__recall_avg_seq_sem

    @property
    def recall_avg_seq_sem_cami1(self):
        return self.__recall_avg_seq_sem_cami1

    @property
    def recall_weighted_bp(self):
        return self.__recall_weighted_bp

    @property
    def recall_weighted_seq(self):
        return self.__recall_weighted_seq

    @property
    def unifrac_bp(self):
        return self.__unifrac_bp

    @property
    def unifrac_seq(self):
        return self.__unifrac_seq

    @percentage_of_assigned_bps.setter
    def percentage_of_assigned_bps(self, percentage_of_assigned_bps):
        self.__percentage_of_assigned_bps = percentage_of_assigned_bps

    @percentage_of_assigned_seqs.setter
    def percentage_of_assigned_seqs(self, percentage_of_assigned_seqs):
        self.__percentage_of_assigned_seqs = percentage_of_assigned_seqs

    @accuracy_bp.setter
    def accuracy_bp(self, accuracy_bp):
        self.__accuracy_bp = accuracy_bp

    @accuracy_seq.setter
    def accuracy_seq(self, accuracy_seq):
        self.__accuracy_seq = accuracy_seq

    @rand_index_bp.setter
    def rand_index_bp(self, rand_index_bp):
        self.__rand_index_bp = rand_index_bp

    @adjusted_rand_index_bp.setter
    def adjusted_rand_index_bp(self, adjusted_rand_index_bp):
        self.__adjusted_rand_index_bp = adjusted_rand_index_bp

    @rand_index_seq.setter
    def rand_index_seq(self, rand_index_seq):
        self.__rand_index_seq = rand_index_seq

    @adjusted_rand_index_seq.setter
    def adjusted_rand_index_seq(self, adjusted_rand_index_seq):
        self.__adjusted_rand_index_seq = adjusted_rand_index_seq

    @precision_avg_bp.setter
    def precision_avg_bp(self, precision_avg_bp):
        self.__precision_avg_bp = precision_avg_bp

    @precision_avg_bp_var.setter
    def precision_avg_bp_var(self, precision_avg_bp_var):
        self.__precision_avg_bp_var = precision_avg_bp_var

    @precision_avg_bp_sem.setter
    def precision_avg_bp_sem(self, precision_avg_bp_sem):
        self.__precision_avg_bp_sem = precision_avg_bp_sem

    @precision_avg_seq.setter
    def precision_avg_seq(self, precision_avg_seq):
        self.__precision_avg_seq = precision_avg_seq

    @precision_avg_seq_sem.setter
    def precision_avg_seq_sem(self, precision_avg_seq_sem):
        self.__precision_avg_seq_sem = precision_avg_seq_sem

    @precision_weighted_bp.setter
    def precision_weighted_bp(self, precision_weighted_bp):
        self.__precision_weighted_bp = precision_weighted_bp

    @precision_weighted_seq.setter
    def precision_weighted_seq(self, precision_weighted_seq):
        self.__precision_weighted_seq = precision_weighted_seq

    @recall_avg_bp.setter
    def recall_avg_bp(self, recall_avg_bp):
        self.__recall_avg_bp = recall_avg_bp

    @recall_avg_bp_cami1.setter
    def recall_avg_bp_cami1(self, recall_avg_bp_cami1):
        self.__recall_avg_bp_cami1 = recall_avg_bp_cami1

    @recall_avg_bp_var.setter
    def recall_avg_bp_var(self, recall_avg_bp_var):
        self.__recall_avg_bp_var = recall_avg_bp_var

    @recall_avg_bp_var_cami1.setter
    def recall_avg_bp_var_cami1(self, recall_avg_bp_var_cami1):
        self.__recall_avg_bp_var_cami1 = recall_avg_bp_var_cami1

    @recall_avg_bp_sem.setter
    def recall_avg_bp_sem(self, recall_avg_bp_sem):
        self.__recall_avg_bp_sem = recall_avg_bp_sem

    @recall_avg_bp_sem_cami1.setter
    def recall_avg_bp_sem_cami1(self, recall_avg_bp_sem_cami1):
        self.__recall_avg_bp_sem_cami1 = recall_avg_bp_sem_cami1

    @recall_avg_seq.setter
    def recall_avg_seq(self, recall_avg_seq):
        self.__recall_avg_seq = recall_avg_seq

    @recall_avg_seq_cami1.setter
    def recall_avg_seq_cami1(self, recall_avg_seq_cami1):
        self.__recall_avg_seq_cami1 = recall_avg_seq_cami1

    @recall_avg_seq_sem.setter
    def recall_avg_seq_sem(self, recall_avg_seq_sem):
        self.__recall_avg_seq_sem = recall_avg_seq_sem

    @recall_avg_seq_sem_cami1.setter
    def recall_avg_seq_sem_cami1(self, recall_avg_seq_sem_cami1):
        self.__recall_avg_seq_sem_cami1 = recall_avg_seq_sem_cami1

    @recall_weighted_bp.setter
    def recall_weighted_bp(self, recall_weighted_bp):
        self.__recall_weighted_bp = recall_weighted_bp

    @recall_weighted_seq.setter
    def recall_weighted_seq(self, recall_weighted_seq):
        self.__recall_weighted_seq = recall_weighted_seq

    @unifrac_bp.setter
    def unifrac_bp(self, unifrac_bp):
        self.__unifrac_bp = unifrac_bp

    @unifrac_seq.setter
    def unifrac_seq(self, unifrac_seq):
        self.__unifrac_seq = unifrac_seq

    @staticmethod
    def compute_rand_index(confusion_df, col_name, gs_col_name, field):
        def choose2(n):
            return math.comb(n, 2)

        bin_mapping_comb = sum(confusion_df[field].apply(choose2))
        bin_comb = sum(confusion_df.groupby(col_name).agg({field: 'sum'})[field].apply(choose2))
        mapping_comb = sum(confusion_df.groupby(gs_col_name).agg({field: 'sum'})[field].apply(choose2))
        num_bp_comb = choose2(sum(confusion_df[field]))

        rand_index = ((num_bp_comb - bin_comb - mapping_comb + 2 * bin_mapping_comb) / num_bp_comb) if num_bp_comb != 0 else .0

        temp = (bin_comb * (mapping_comb / num_bp_comb)) if num_bp_comb != 0 else .0
        ret = bin_mapping_comb - temp
        denominator = (((bin_comb + mapping_comb) / 2.0) - temp)
        adjusted_rand_index = (ret / denominator) if denominator != 0 else .0

        return rand_index, adjusted_rand_index

    def get_ordered_dict(self):
        def f1_score(metric1, metric2):
            if metric1 + metric2 > 0:
                return 2 * metric1 * metric2 / (metric1 + metric2)
            else:
                return np.nan

        metrics_dict = {metric.split("__")[1]: value for metric, value in self.__dict__.items()}
        metrics_dict['f1_score_bp'] = f1_score(self.__precision_avg_bp, self.__recall_avg_bp)
        metrics_dict['f1_score_seq'] = f1_score(self.__precision_avg_seq, self.__recall_avg_seq)
        metrics_dict['f1_score_bp_cami1'] = f1_score(self.__precision_avg_bp, self.__recall_avg_bp_cami1)
        metrics_dict['f1_score_seq_cami1'] = f1_score(self.__precision_avg_seq, self.__recall_avg_seq_cami1)
        metrics_dict['f1_score_per_bp'] = f1_score(self.__precision_weighted_bp, self.__recall_weighted_bp)
        metrics_dict['f1_score_per_seq'] = f1_score(self.__precision_weighted_seq, self.__recall_weighted_seq)
        metrics_dict['misclassification_bp'] = 1 - self.__precision_weighted_bp
        metrics_dict['misclassification_seq'] = 1 - self.__precision_weighted_seq

        return OrderedDict([(utils_labels.TOOL, None),
                            (utils_labels.BINNING_TYPE, None),
                            (utils_labels.SAMPLE, None),
                            (utils_labels.RANK, None)] +
                            [(k, [v]) for k, v in metrics_dict.items()])


class Query(ABC):
    def __init__(self, label, sample_id, options, metadata, is_gs=False):
        self.__label = label
        self.__sample_id = sample_id
        self.__gold_standard = None
        self.__precision_df = pd.DataFrame()
        self.__recall_df = pd.DataFrame()
        self.__heatmap_sdf = None
        self.__metrics = None
        self.__metrics_filtered = None
        self.__options = options
        self.__metadata = metadata
        self.__is_gs = is_gs
        self.__eval_success = False

    @property
    def label(self):
        return self.__label

    @property
    def sample_id(self):
        return self.__sample_id

    @property
    def gold_standard(self):
        return self.__gold_standard

    @property
    def precision_df(self):
        return self.__precision_df

    @property
    def recall_df(self):
        return self.__recall_df

    @property
    def heatmap_sdf(self):
        return self.__heatmap_sdf

    @property
    def metrics(self):
        return self.__metrics

    @property
    def metrics_filtered(self):
        return self.__metrics_filtered

    @property
    def options(self):
        return self.__options

    @property
    def metadata(self):
        return self.__metadata

    @property
    def is_gs(self):
        return self.__is_gs

    @property
    def eval_success(self):
        return self.__eval_success

    @label.setter
    def label(self, label):
        self.__label = label

    @sample_id.setter
    def sample_id(self, sample_id):
        self.__sample_id = sample_id

    @gold_standard.setter
    def gold_standard(self, gold_standard):
        self.__gold_standard = gold_standard

    @precision_df.setter
    def precision_df(self, precision_df):
        self.__precision_df = precision_df

    @recall_df.setter
    def recall_df(self, recall_df):
        self.__recall_df = recall_df

    @heatmap_sdf.setter
    def heatmap_sdf(self, heatmap_sdf):
        self.__heatmap_sdf = heatmap_sdf

    @metrics.setter
    def metrics(self, metrics):
        self.__metrics = metrics

    @metrics_filtered.setter
    def metrics_filtered(self, metrics_filtered):
        self.__metrics_filtered = metrics_filtered

    @options.setter
    def options(self, options):
        self.__options = options

    @metadata.setter
    def metadata(self, metadata):
        self.__metadata = metadata

    @eval_success.setter
    def eval_success(self, eval_success):
        self.__eval_success = eval_success

    @abstractmethod
    def compute_metrics(self):
        pass

    def compute_unifrac(self, all_bins):
        return None, None

    def plot(self):
        return


class GenomeQuery(Query):
    binning_type = 'genome'

    def __init__(self, label, sample_id, options, metadata, is_gs=False):
        super().__init__(label, sample_id, options, metadata, is_gs)
        self.__recall_df_cami1 = pd.DataFrame()
        self.metrics = Metrics()

    @property
    def df(self):
        query_df = load_data.load_sample(self.metadata).drop_duplicates(['SEQUENCEID', 'BINID'])
        if self.options.min_length and self.is_gs:
            query_df = query_df[query_df['LENGTH'] >= self.options.min_length]
        if self.is_gs:
            return query_df[['SEQUENCEID', 'BINID', 'LENGTH']]
        else:
            return query_df[['SEQUENCEID', 'BINID']]

    @property
    def gold_standard_data(self):
        return self.gold_standard.df[['SEQUENCEID', 'BINID', 'LENGTH']].rename(columns={'LENGTH': 'seq_length', 'BINID': 'genome_id'})

    @property
    def recall_df_cami1(self):
        return self.__recall_df_cami1

    @recall_df_cami1.setter
    def recall_df_cami1(self, recall_df_cami1):
        self.__recall_df_cami1 = recall_df_cami1

    def get_metrics_df(self):
        metrics_dict = self.metrics.get_ordered_dict()
        metrics_dict[utils_labels.TOOL] = self.label
        metrics_dict[utils_labels.BINNING_TYPE] = self.binning_type
        metrics_dict[utils_labels.RANK] = 'NA'
        metrics_dict[utils_labels.SAMPLE] = self.sample_id
        return pd.DataFrame(metrics_dict)

    def compute_metrics(self, gs_df):
        if self.label == utils_labels.GS and (self.options.only_taxonomic_queries or self.options.skip_gs):
            return False
        logging.getLogger('amber').info('Evaluating {}, sample {}, genome binning'.format(self.label, self.sample_id))

        query_df = self.df
        condition = query_df['SEQUENCEID'].isin(gs_df['SEQUENCEID'])
        if ~condition.all():
            logging.getLogger('amber').warning("{} sequences in {} not found in the gold standard.".format(query_df[~condition]['SEQUENCEID'].nunique(), self.label))
            query_df = query_df[condition]

        query_w_length = pd.merge(query_df, gs_df.drop_duplicates('SEQUENCEID'), on='SEQUENCEID', sort=False)

        query_w_length_no_dups = query_w_length.drop_duplicates('SEQUENCEID')
        gs_df_no_dups = gs_df.drop_duplicates('SEQUENCEID')
        self.metrics.percentage_of_assigned_bps = query_w_length_no_dups['seq_length'].sum() / gs_df_no_dups['seq_length'].sum()
        self.metrics.percentage_of_assigned_seqs = query_w_length_no_dups.shape[0] / gs_df_no_dups['SEQUENCEID'].shape[0]
        del query_w_length_no_dups

        # confusion table possibly with the same sequences in multiple bins
        query_w_length_mult_seqs = query_df.reset_index().merge(gs_df, on='SEQUENCEID', sort=False)

        if query_w_length.shape[0] < query_w_length_mult_seqs.shape[0]:
            query_w_length_mult_seqs.drop_duplicates(['index', 'genome_id'], inplace=True)
            confusion_df = query_w_length_mult_seqs.groupby(['BINID', 'genome_id'], sort=False).agg({'seq_length': 'sum', 'SEQUENCEID': 'count'}).rename(columns={'seq_length': 'genome_length', 'SEQUENCEID': 'genome_seq_counts'})

            most_abundant_genome_df = confusion_df.loc[confusion_df.groupby('BINID', sort=False)['genome_length'].idxmax()]
            most_abundant_genome_df = most_abundant_genome_df.reset_index()[['BINID', 'genome_id']]

            matching_genomes_df = pd.merge(query_w_length_mult_seqs, most_abundant_genome_df, on=['BINID', 'genome_id']).set_index('index')
            query_w_length_mult_seqs.set_index('index', inplace=True)
            difference_df = query_w_length_mult_seqs.drop(matching_genomes_df.index).groupby(['index'], sort=False).first()
            query_w_length = pd.concat([matching_genomes_df, difference_df])

            # Modify gs such that multiple binnings of the same sequence are not required
            matching_genomes_df = pd.merge(gs_df, query_w_length[['SEQUENCEID', 'genome_id']], on=['SEQUENCEID', 'genome_id'])
            matching_genomes_df = matching_genomes_df[['SEQUENCEID', 'genome_id', 'seq_length']].drop_duplicates(['SEQUENCEID', 'genome_id'])
            condition = gs_df_no_dups['SEQUENCEID'].isin(matching_genomes_df['SEQUENCEID'])
            difference_df = gs_df_no_dups[~condition]
            gs_df = pd.concat([difference_df, matching_genomes_df])

            # query_w_length_mult_seqs.reset_index(inplace=True)
            # query_w_length_mult_seqs = pd.merge(query_w_length_mult_seqs, most_abundant_genome_df, on=['BINID'])
            # grouped = query_w_length_mult_seqs.groupby(['index'], sort=False, as_index=False)
            # query_w_length = grouped.apply(lambda x: x[x['genome_id_x'] == x['genome_id_y'] if any(x['genome_id_x'] == x['genome_id_y']) else len(x) * [True]])
            # query_w_length = query_w_length.groupby(['index'], sort=False).first().drop(columns='genome_id_y').rename(columns={'genome_id_x': 'genome_id'})

        del query_w_length_mult_seqs

        confusion_df = query_w_length.groupby(['BINID', 'genome_id'], sort=False).agg({'seq_length': 'sum', 'SEQUENCEID': 'count'}).rename(columns={'seq_length': 'genome_length', 'SEQUENCEID': 'genome_seq_counts'})

        self.metrics.rand_index_bp, self.metrics.adjusted_rand_index_bp = Metrics.compute_rand_index(confusion_df, 'BINID', 'genome_id', 'genome_length')
        self.metrics.rand_index_seq, self.metrics.adjusted_rand_index_seq = Metrics.compute_rand_index(confusion_df, 'BINID', 'genome_id', 'genome_seq_counts')

        most_abundant_genome_df = confusion_df.loc[confusion_df.groupby('BINID', sort=False)['genome_length'].idxmax()].reset_index().set_index('BINID')

        query_w_length['seq_length_mean'] = query_w_length['seq_length']

        precision_df = query_w_length.groupby('BINID', sort=False).agg({'seq_length': 'sum', 'seq_length_mean': 'mean', 'SEQUENCEID': 'count'}).rename(columns={'seq_length': 'total_length', 'SEQUENCEID': 'total_seq_counts'})
        precision_df = pd.merge(precision_df, most_abundant_genome_df, on='BINID')
        precision_df.rename(columns={'genome_length': 'tp_length', 'genome_seq_counts': 'tp_seq_counts'}, inplace=True)
        precision_df['precision_bp'] = precision_df['tp_length'] / precision_df['total_length']
        precision_df['precision_seq'] = precision_df['tp_seq_counts'] / precision_df['total_seq_counts']

        if self.options.filter_tail_percentage:
            precision_df['total_length_pct'] = precision_df['total_length'] / precision_df['total_length'].sum()
            precision_df.sort_values(by='total_length', inplace=True)
            precision_df['cumsum_length_pct'] = precision_df['total_length_pct'].cumsum(axis=0)
            precision_df['precision_bp'] = precision_df['precision_bp'].mask(precision_df['cumsum_length_pct'] <= self.options.filter_tail_percentage / 100)
            precision_df['precision_seq'] = precision_df['precision_seq'].mask(precision_df['precision_bp'].isna())
            precision_df.drop(columns=['cumsum_length_pct', 'total_length_pct'], inplace=True)
        if self.options.genome_to_unique_common:
            precision_df = precision_df[~precision_df['genome_id'].isin(self.options.genome_to_unique_common)]

        self.metrics.precision_avg_bp = precision_df['precision_bp'].mean()
        self.metrics.precision_avg_bp_sem = precision_df['precision_bp'].sem()
        self.metrics.precision_avg_bp_var = precision_df['precision_bp'].var()
        self.metrics.precision_avg_seq = precision_df['precision_seq'].mean()
        self.metrics.precision_avg_seq_sem = precision_df['precision_seq'].sem()

        def safe_divide(x, y):
            return x / y if y else np.nan

        self.metrics.precision_weighted_bp = safe_divide(precision_df['tp_length'].sum(), precision_df['total_length'].sum())
        self.metrics.precision_weighted_seq = safe_divide(precision_df['tp_seq_counts'].sum(), precision_df['total_seq_counts'].sum())

        genome_sizes_df = gs_df.groupby('genome_id', sort=False).agg({'seq_length': 'sum', 'SEQUENCEID': 'count'}).rename(columns={'seq_length': 'length_gs', 'SEQUENCEID': 'seq_counts_gs'})
        precision_df = precision_df.reset_index().join(genome_sizes_df, on='genome_id', how='left', sort=False).set_index('BINID')
        precision_df['recall_bp'] = precision_df['tp_length'] / precision_df['length_gs']
        precision_df['recall_seq'] = precision_df['tp_seq_counts'] / precision_df['seq_counts_gs']
        precision_df['rank'] = 'NA'

        recall_df = confusion_df.loc[confusion_df.groupby('genome_id', sort=False)['genome_length'].idxmax()]
        recall_df = recall_df.reset_index().join(genome_sizes_df, on='genome_id', how='right', sort=False).set_index('BINID')
        recall_df.fillna({'genome_length': 0, 'genome_seq_counts': 0}, inplace=True)
        recall_df['recall_bp'] = recall_df['genome_length'] / recall_df['length_gs']
        recall_df['recall_seq'] = recall_df['genome_seq_counts'] / recall_df['seq_counts_gs']

        recall_df = recall_df.join(precision_df[['total_length', 'seq_length_mean']], how='left', sort=False)

        if self.options.genome_to_unique_common:
            recall_df = recall_df[~recall_df['genome_id'].isin(self.options.genome_to_unique_common)]

        self.metrics.recall_avg_bp = recall_df['recall_bp'].mean()
        self.metrics.recall_avg_bp_var = recall_df['recall_bp'].var()
        self.metrics.recall_avg_bp_sem = recall_df['recall_bp'].sem()
        self.metrics.recall_avg_seq = recall_df['recall_seq'].mean()
        self.metrics.recall_avg_seq_sem = recall_df['recall_seq'].sem()
        self.metrics.recall_weighted_bp = recall_df['genome_length'].sum() / recall_df['length_gs'].sum()
        self.metrics.recall_weighted_seq = recall_df['genome_seq_counts'].sum() / recall_df['seq_counts_gs'].sum()

        # Compute recall as in CAMI 1
        unmapped_genomes = set(gs_df['genome_id'].unique()) - set(precision_df['genome_id'].unique())
        if self.options.genome_to_unique_common:
            unmapped_genomes -= set(self.options.genome_to_unique_common)
        num_unmapped_genomes = len(unmapped_genomes)
        prec_copy = precision_df.reset_index()
        if num_unmapped_genomes:
            numeric_cols = prec_copy.select_dtypes('number').columns
            prec_copy = prec_copy.reindex(prec_copy.index.tolist() + list(range(len(prec_copy), len(prec_copy) + num_unmapped_genomes))).fillna(dict.fromkeys(numeric_cols, 0))
        self.metrics.recall_avg_bp_cami1 = prec_copy['recall_bp'].mean()
        self.metrics.recall_avg_seq_cami1 = prec_copy['recall_seq'].mean()
        self.metrics.recall_avg_bp_sem_cami1 = prec_copy['recall_bp'].sem()
        self.metrics.recall_avg_seq_sem_cami1 = prec_copy['recall_seq'].sem()
        self.metrics.recall_avg_bp_var_cami1 = prec_copy['recall_bp'].var()
        self.recall_df_cami1 = prec_copy
        # End Compute recall as in CAMI 1

        self.metrics.accuracy_bp = precision_df['tp_length'].sum() / recall_df['length_gs'].sum()
        self.metrics.accuracy_seq = precision_df['tp_seq_counts'].sum() / recall_df['seq_counts_gs'].sum()

        self.precision_df = precision_df.sort_values(by=['recall_bp'], axis=0, ascending=False)
        self.precision_df[utils_labels.TOOL] = self.label
        self.precision_df['sample_id'] = self.sample_id
        self.recall_df = recall_df

        self.heatmap_sdf = precision_recall_per_bin.transform_confusion_matrix2(query_w_length, confusion_df, precision_df, gs_df, log_scale=True)

        self.eval_success = True

    @staticmethod
    def calc_num_recovered_genomes(pd_bins, min_completeness, max_contamination):
        counts_list = []
        for (sample_id, tool), pd_group in pd_bins.groupby(['sample_id', utils_labels.TOOL]):
            for x in itertools.product(min_completeness, max_contamination):
                count = pd_group[(pd_group['recall_bp'] > x[0]) & (pd_group['precision_bp'] > (1 - x[1]))].shape[0]
                counts_list.append((sample_id, tool, '> ' + str(int(x[0] * 100)) + '% completeness', '< ' + str(int(x[1] * 100)) + '%', count))

        pd_counts = pd.DataFrame(counts_list, columns=[utils_labels.SAMPLE, utils_labels.TOOL, 'Completeness', 'Contamination', 'count'])
        pd_counts = pd.pivot_table(pd_counts, values='count', index=[utils_labels.SAMPLE, utils_labels.TOOL, 'Contamination'], columns=['Completeness']).reset_index()
        return pd_counts

    def plot_precision_vs_bin_size(self):
        fig, axs = plt.subplots(figsize=(5, 4.5))
        df_sorted = self.precision_df[['total_length', 'precision_bp']].sort_values(by=['total_length'])

        axs.scatter(np.log(df_sorted['total_length']), df_sorted['precision_bp'], marker='o')
        window = int(df_sorted.shape[0] / 50) if df_sorted.shape[0] > 100 else int(df_sorted.shape[0] / 10)
        rolling_mean = df_sorted['precision_bp'].rolling(window=window, min_periods=int(window/2)).mean()
        axs.plot(np.log(df_sorted['total_length']), rolling_mean, color='orange')

        axs.set_xlim([None, np.log(df_sorted['total_length'].max())])
        axs.set_ylim([0.0, 1.0])
        axs.set_title(self.label, fontsize=12)
        plt.ylabel('Purity per bin (%)', fontsize=12)
        plt.xlabel('Bin size [log(# bp)]', fontsize=12)
        fig.savefig(os.path.join(self.options.output_dir, 'genome', self.label, 'purity_vs_bin_size_' + self.sample_id + '.png'), dpi=200, format='png', bbox_inches='tight')
        fig.savefig(os.path.join(self.options.output_dir, 'genome', self.label, 'purity_vs_bin_size_' + self.sample_id + '.pdf'), dpi=200, format='pdf', bbox_inches='tight')
        plt.close(fig)

    def plot_recall_vs_genome_size(self):
        fig, axs = plt.subplots(figsize=(5, 4.5))
        df_sorted = self.recall_df[['total_length', 'recall_bp']].sort_values(by=['total_length'])

        axs.scatter(np.log(df_sorted['total_length']), df_sorted['recall_bp'], marker='o')
        window = int(df_sorted.shape[0] / 50) if df_sorted.shape[0] > 100 else int(df_sorted.shape[0] / 10)
        rolling_mean = df_sorted['recall_bp'].rolling(window=window, min_periods=int(window / 2)).mean()
        axs.plot(np.log(df_sorted['total_length']), rolling_mean, color='orange')

        axs.set_xlim([None, np.log(self.recall_df['total_length'].max())])
        axs.set_ylim([0.0, 1.0])
        axs.set_title(self.label, fontsize=12)
        plt.ylabel('Completeness per genome (%)', fontsize=12)
        plt.xlabel('Genome size [log(# bp)]', fontsize=12)
        fig.savefig(os.path.join(self.options.output_dir, 'genome', self.label, 'completeness_vs_genome_size_' + self.sample_id + '.png'), dpi=200, format='png', bbox_inches='tight')
        fig.savefig(os.path.join(self.options.output_dir, 'genome', self.label, 'completeness_vs_genome_size_' + self.sample_id + '.pdf'), dpi=200, format='pdf', bbox_inches='tight')
        plt.close(fig)

    def plot_heat_maps(self):
        if self.label == utils_labels.GS:
            return
        plots.plot_heatmap(self.heatmap_sdf, self.sample_id, self.options.output_dir, self.label, log_scale=True)

    def plot(self):
        if self.label == utils_labels.GS:
            return
        self.plot_precision_vs_bin_size()
        self.plot_recall_vs_genome_size()
        self.plot_heat_maps()


class Prediction:
    def __init__(self):
        pass


class TaxonomicQuery(Query):
    binning_type = 'taxonomic'

    def __init__(self, label, sample_id, options, metadata, taxonomy_df, is_gs=False):
        super().__init__(label, sample_id, options, metadata, is_gs)
        self.__rank_to_df = None
        self.metrics = defaultdict()
        if self.options.filter_tail_percentage:
            self.metrics_filtered = defaultdict()
        self.__profile = None
        self.__profile_filtered = None
        self.__taxonomy_df = taxonomy_df

    @property
    def rank_to_df(self):
        query_df = load_data.load_sample(self.metadata)
        if self.options.min_length and self.is_gs:
            query_df = query_df[query_df['LENGTH'] >= self.options.min_length]
        rank_to_df = load_data.get_rank_to_df(query_df, self.taxonomy_df, self.label, self.is_gs)
        del query_df
        return rank_to_df

    @property
    def gold_standard_data(self):
        return self.gold_standard.rank_to_df

    @property
    def profile(self):
        if self.__profile:
            return self.__profile
        self.__profile = self._create_profile(all_bins=True)
        return self.__profile

    @property
    def profile_filtered(self):
        if self.__profile_filtered:
            return self.__profile_filtered
        self.__profile_filtered = self._create_profile(all_bins=False)
        return self.__profile_filtered

    @property
    def taxonomy_df(self):
        return self.__taxonomy_df

    @rank_to_df.setter
    def rank_to_df(self, rank_to_df):
        self.__rank_to_df = rank_to_df

    def _create_profile(self, all_bins):
        if self.recall_df.empty:
            return [], []

        profile_bp = []
        profile_seq = []
        if all_bins:
            recall_df = self.recall_df
        else:
            recall_df = self.recall_df[~self.recall_df['filtered']]
        for index, row in recall_df.iterrows():
            prediction_bp = Prediction()
            prediction_bp.taxid = str(int(row['TAXID']))
            prediction_bp.rank = row['rank']
            prediction_bp.percentage = row['tp_length']
            taxpath = self.taxonomy_df.loc[row['TAXID']][load_ncbi_taxinfo.RANKS].values
            taxpath = '|'.join(map(lambda v: '' if pd.isnull(v) else str(v), taxpath)).rstrip('|')
            prediction_bp.taxpath = taxpath
            prediction_bp.taxpathsn = None
            profile_bp.append(prediction_bp)

            prediction_seq = Prediction()
            prediction_seq.taxid = prediction_bp.taxid
            prediction_seq.rank = prediction_bp.rank
            prediction_seq.percentage = row['tp_seq_counts']
            prediction_seq.taxpath = prediction_bp.taxpath
            prediction_seq.taxpathsn = None
            profile_seq.append(prediction_seq)
        return profile_bp, profile_seq

    def compute_unifrac(self, all_bins):
        if all_bins:
            pf_profile_bp = pf.Profile(profile=self.profile[0])
            pf_profile_seq = pf.Profile(profile=self.profile[1])
        else:
            pf_profile_bp = pf.Profile(profile=self.profile_filtered[0])
            pf_profile_seq = pf.Profile(profile=self.profile_filtered[1])
        gs_pf_profile_bp = pf.Profile(profile=self.gold_standard.profile[0])
        gs_pf_profile_seq = pf.Profile(profile=self.gold_standard.profile[1])
        return uf.compute_unifrac(gs_pf_profile_bp, pf_profile_bp), uf.compute_unifrac(gs_pf_profile_seq, pf_profile_seq)

    def get_metrics_df(self):
        allranks_metrics_df = pd.DataFrame()
        for rank in self.metrics:
            metrics_dict = self.metrics[rank].get_ordered_dict()
            rank_metrics_df = pd.DataFrame(metrics_dict)

            if self.metrics_filtered:
                rank_metrics_df = pd.DataFrame(metrics_dict) \
                    .drop(columns=[utils_labels.TOOL, utils_labels.BINNING_TYPE, utils_labels.SAMPLE, utils_labels.RANK]) \
                    .add_suffix(utils_labels.UNFILTERED)
                metrics_dict = self.metrics_filtered[rank].get_ordered_dict()
                rank_metrics_df_filtered = pd.DataFrame(metrics_dict)
                rank_metrics_df = pd.concat([rank_metrics_df_filtered, rank_metrics_df], axis=1)

            rank_metrics_df[utils_labels.TOOL] = self.label
            rank_metrics_df[utils_labels.BINNING_TYPE] = self.binning_type
            rank_metrics_df[utils_labels.RANK] = rank
            rank_metrics_df[utils_labels.SAMPLE] = self.sample_id

            allranks_metrics_df = pd.concat([allranks_metrics_df, rank_metrics_df], ignore_index=True, sort=True)
        return allranks_metrics_df

    def compute_metrics_per_rank(self, rank_to_df, gs_rank_to_df, rank):
        if rank not in gs_rank_to_df:
            return
        logging.getLogger('amber').info('Evaluating {}, sample {}, taxonomic binning, {}'.format(self.label, self.sample_id, rank))

        self.metrics[rank] = Metrics()
        gs_df = gs_rank_to_df[rank]

        query_df = rank_to_df[rank][['SEQUENCEID', 'TAXID']]
        condition = query_df['SEQUENCEID'].isin(gs_df['SEQUENCEID'])
        if ~condition.all():
            logging.getLogger('amber').warning("{} sequences in {} not found in the gold standard.".format(query_df[~condition]['SEQUENCEID'].nunique(), self.label))
            query_df = query_df[condition]

        query_w_length_df = pd.merge(query_df, gs_df.rename(columns={'TAXID': 'true_taxid'}).reset_index(), on='SEQUENCEID', sort=False)

        confusion_df = query_w_length_df.groupby(['TAXID', 'true_taxid'], sort=False).agg({'LENGTH': 'sum', 'SEQUENCEID': 'count'}).rename(columns={'LENGTH': 'tax_length', 'SEQUENCEID': 'tax_seq_counts'})
        self.metrics[rank].rand_index_bp, self.metrics[rank].adjusted_rand_index_bp = Metrics.compute_rand_index(confusion_df, 'TAXID', 'true_taxid', 'tax_length')
        self.metrics[rank].rand_index_seq, self.metrics[rank].adjusted_rand_index_seq = Metrics.compute_rand_index(confusion_df, 'TAXID', 'true_taxid', 'tax_seq_counts')

        query_w_length_df = query_w_length_df[['SEQUENCEID', 'TAXID', 'LENGTH']]

        self.metrics[rank].percentage_of_assigned_bps = query_w_length_df['LENGTH'].sum() / gs_df['LENGTH'].sum()
        self.metrics[rank].percentage_of_assigned_seqs = query_w_length_df.shape[0] / gs_df.shape[0]

        true_positives_df = pd.merge(query_df, gs_df, on=['SEQUENCEID', 'TAXID'], sort=False)
        true_positives_df = true_positives_df.groupby('TAXID', sort=False).agg({'LENGTH': 'sum', 'SEQUENCEID': 'count'}).rename(columns={'LENGTH': 'tp_length', 'SEQUENCEID': 'tp_seq_counts'})

        tp_fp_fn_df = query_w_length_df.groupby('TAXID', sort=False).agg({'LENGTH': 'sum', 'SEQUENCEID': 'count'}).rename(columns={'LENGTH': 'total_length', 'SEQUENCEID': 'total_seq_counts'})
        tp_fp_fn_df = pd.merge(true_positives_df, tp_fp_fn_df, on=['TAXID'], how='outer', sort=False)
        tax_sizes_df = gs_df.groupby('TAXID', sort=False).agg({'LENGTH': 'sum', 'SEQUENCEID': 'count'}).rename(columns={'LENGTH': 'length_gs', 'SEQUENCEID': 'seq_counts_gs'})
        tp_fp_fn_df = tp_fp_fn_df.reset_index().join(tax_sizes_df, on='TAXID', how='outer', sort=False).set_index('TAXID').fillna(0).astype('int64')

        tp_fp_fn_df['precision_bp'] = tp_fp_fn_df['tp_length'] / tp_fp_fn_df['total_length']
        tp_fp_fn_df['precision_seq'] = tp_fp_fn_df['tp_seq_counts'] / tp_fp_fn_df['total_seq_counts']
        tp_fp_fn_df['recall_bp'] = tp_fp_fn_df['tp_length'] / tp_fp_fn_df['length_gs']
        tp_fp_fn_df['recall_seq'] = tp_fp_fn_df['tp_seq_counts'] / tp_fp_fn_df['seq_counts_gs']
        tp_fp_fn_df['rank'] = rank

        length_gs_sum = tp_fp_fn_df['length_gs'].sum()
        seq_counts_gs_sum = tp_fp_fn_df['seq_counts_gs'].sum()

        def set_values(metric_obj, df):
            tp_length_sum = df['tp_length'].sum()
            tp_seq_counts_sum = df['tp_seq_counts'].sum()

            metric_obj.precision_avg_bp = df['precision_bp'].mean()
            metric_obj.precision_avg_bp_sem = df['precision_bp'].sem()
            metric_obj.precision_avg_seq = df['precision_seq'].mean()
            metric_obj.precision_avg_seq_sem = df['precision_seq'].sem()
            metric_obj.precision_weighted_bp = tp_length_sum / df['total_length'].sum()
            metric_obj.precision_weighted_seq = tp_seq_counts_sum / df['total_seq_counts'].sum()

            metric_obj.recall_avg_bp = df['recall_bp'].mean()
            metric_obj.recall_avg_bp_sem = df['recall_bp'].sem()
            metric_obj.recall_avg_seq = df['recall_seq'].mean()
            metric_obj.recall_avg_seq_sem = df['recall_seq'].sem()
            metric_obj.recall_weighted_bp = tp_length_sum / length_gs_sum
            metric_obj.recall_weighted_seq = tp_seq_counts_sum / seq_counts_gs_sum

            metric_obj.accuracy_bp = tp_length_sum / length_gs_sum
            metric_obj.accuracy_seq = tp_seq_counts_sum / seq_counts_gs_sum

        set_values(self.metrics[rank], tp_fp_fn_df)

        if self.options.filter_tail_percentage:
            df_cpy = tp_fp_fn_df.copy()
            df_cpy['total_length_pct'] = df_cpy['total_length'] / df_cpy['total_length'].sum()
            df_cpy.sort_values(by='total_length', inplace=True)
            df_cpy['cumsum_length_pct'] = df_cpy['total_length_pct'].cumsum(axis=0)
            nan_rows = df_cpy['cumsum_length_pct'] <= self.options.filter_tail_percentage / 100
            df_cpy['precision_bp'] = df_cpy['precision_bp'].mask(nan_rows)
            df_cpy['precision_seq'] = df_cpy['precision_seq'].mask(nan_rows)
            df_cpy['recall_bp'] = df_cpy['recall_bp'].mask(nan_rows, other=.0)
            df_cpy['recall_seq'] = df_cpy['recall_seq'].mask(nan_rows, other=.0)
            df_cpy['tp_length'] = df_cpy['tp_length'].mask(nan_rows)
            df_cpy['tp_seq_counts'] = df_cpy['tp_seq_counts'].mask(nan_rows)
            df_cpy['total_length'] = df_cpy['total_length'].mask(nan_rows)
            df_cpy['total_seq_counts'] = df_cpy['total_seq_counts'].mask(nan_rows)
            df_cpy.drop(columns=['cumsum_length_pct', 'total_length_pct'], inplace=True)

            self.metrics_filtered[rank] = Metrics()
            set_values(self.metrics_filtered[rank], df_cpy)

            tp_fp_fn_df['filtered'] = nan_rows

            remaining = df_cpy[~nan_rows].index.values
            confusion_df = confusion_df.loc[remaining]
            self.metrics_filtered[rank].rand_index_bp, self.metrics_filtered[rank].adjusted_rand_index_bp = Metrics.compute_rand_index(confusion_df, 'TAXID', 'true_taxid', 'tax_length')
            self.metrics_filtered[rank].rand_index_seq, self.metrics_filtered[rank].adjusted_rand_index_seq = Metrics.compute_rand_index(confusion_df, 'TAXID', 'true_taxid', 'tax_seq_counts')

            query_w_length_df = query_w_length_df.set_index('TAXID').loc[remaining]
            self.metrics_filtered[rank].percentage_of_assigned_bps = query_w_length_df['LENGTH'].sum() / gs_df['LENGTH'].sum()
            self.metrics_filtered[rank].percentage_of_assigned_seqs = query_w_length_df.shape[0] / gs_df.shape[0]
        else:
            tp_fp_fn_df['filtered'] = False

        self.precision_df = pd.concat([self.precision_df, tp_fp_fn_df.reset_index().sort_values(
            by='recall_bp', axis=0, ascending=False)], ignore_index=True, sort=True)
        self.recall_df = self.precision_df

        self.recall_df['name'] = self.recall_df['TAXID'].apply(lambda x: self.taxonomy_df.loc[x]['name'])

    def compute_metrics(self, gs_rank_to_df):
        if self.label == utils_labels.GS and (self.options.only_genome_queries or self.options.skip_gs):
            return False
        rank_to_df = self.rank_to_df
        for rank in rank_to_df:
            self.compute_metrics_per_rank(rank_to_df, gs_rank_to_df, rank)
        del rank_to_df

        unifrac_bp, unifrac_seq = self.compute_unifrac(all_bins=True)
        for rank in self.metrics:
            self.metrics[rank].unifrac_bp = unifrac_bp
            self.metrics[rank].unifrac_seq = unifrac_seq
        if self.options.filter_tail_percentage:
            unifrac_bp, unifrac_seq = self.compute_unifrac(all_bins=False)
            for rank in self.metrics:
                self.metrics_filtered[rank].unifrac_bp = unifrac_bp
                self.metrics_filtered[rank].unifrac_seq = unifrac_seq

        self.precision_df[utils_labels.TOOL] = self.label
        self.precision_df['sample_id'] = self.sample_id
        self.eval_success = True


class Options:
    def __init__(self, filter_tail_percentage=0, genome_to_unique_common=None, filter_keyword=None, min_length=0,
                 rank_as_genome_binning=None, output_dir=None, min_completeness=None, max_contamination=None,
                 ncbi_dir=None, skip_gs=False):
        self.__filter_tail_percentage = float(filter_tail_percentage) if filter_tail_percentage else .0
        self.__genome_to_unique_common = genome_to_unique_common
        self.__filter_keyword = filter_keyword
        self.__min_length = int(min_length) if min_length else 0
        if rank_as_genome_binning and rank_as_genome_binning not in load_ncbi_taxinfo.RANKS:
            exit("Not a valid rank to assess taxonomic binning as genome binning (option --rank_as_genome_binning): " + rank_as_genome_binning)
        self.__rank_as_genome_binning = rank_as_genome_binning
        self.__only_genome_queries = True
        self.__only_taxonomic_queries = True
        self.__output_dir = output_dir
        if min_completeness:
            self.__min_completeness = [int(x.strip()) / 100.0 for x in min_completeness.split(',')]
        else:
            self.__min_completeness = [.5, .7, .9]
        if max_contamination:
            self.__max_contamination = [int(x.strip()) / 100.0 for x in max_contamination.split(',')]
        else:
            self.__max_contamination = [.1, .05]
        self.__skip_gs = skip_gs
        self.__ncbi_dir = ncbi_dir

    @property
    def filter_tail_percentage(self):
        return self.__filter_tail_percentage

    @property
    def genome_to_unique_common(self):
        return self.__genome_to_unique_common

    @property
    def filter_keyword(self):
        return self.__filter_keyword

    @property
    def min_length(self):
        return self.__min_length

    @property
    def rank_as_genome_binning(self):
        return self.__rank_as_genome_binning

    @property
    def only_genome_queries(self):
        return self.__only_genome_queries

    @property
    def only_taxonomic_queries(self):
        return self.__only_taxonomic_queries

    @property
    def output_dir(self):
        return self.__output_dir

    @property
    def min_completeness(self):
        return self.__min_completeness

    @property
    def max_contamination(self):
        return self.__max_contamination

    @property
    def ncbi_dir(self):
        return self.__ncbi_dir

    @property
    def skip_gs(self):
        return self.__skip_gs

    @filter_tail_percentage.setter
    def filter_tail_percentage(self, filter_tail_percentage):
        self.__filter_tail_percentage = filter_tail_percentage

    @genome_to_unique_common.setter
    def genome_to_unique_common(self, genome_to_unique_common):
        self.__genome_to_unique_common = genome_to_unique_common

    @filter_keyword.setter
    def filter_keyword(self, filter_keyword):
        self.__filter_keyword = filter_keyword

    @min_length.setter
    def min_length(self, min_length):
        self.__min_length = min_length

    @rank_as_genome_binning.setter
    def rank_as_genome_binning(self, rank_as_genome_binning):
        self.__rank_as_genome_binning = rank_as_genome_binning

    @only_genome_queries.setter
    def only_genome_queries(self, only_genome_queries):
        self.__only_genome_queries = only_genome_queries

    @only_taxonomic_queries.setter
    def only_taxonomic_queries(self, only_taxonomic_queries):
        self.__only_taxonomic_queries = only_taxonomic_queries

    @output_dir.setter
    def output_dir(self, output_dir):
        self.__output_dir = output_dir

    @min_completeness.setter
    def min_completeness(self, min_completeness):
        self.__min_completeness = min_completeness

    @max_contamination.setter
    def max_contamination(self, max_contamination):
        self.__max_contamination = max_contamination

    @ncbi_dir.setter
    def ncbi_dir(self, ncbi_dir):
        self.__ncbi_dir = ncbi_dir

    @skip_gs.setter
    def skip_gs(self, skip_gs):
        self.__skip_gs = skip_gs
